#include "header.h"

module toephssrs

  implicit none
  include 'fftw3.f'

  complex(MYKIND),parameter  :: ZERO = (0.0,0.0), ONE = (1.0,0.0)
  real(MYKIND),   parameter  :: PI =  3.141592653589793 !atan(1.0_16)*4

  type matrix_wrapper
     integer                     :: n
     complex(MYKIND),allocatable :: cv(:), rv(:), d(:), x(:)
     complex(MYKIND),allocatable :: u(:,:), v(:,:)
  end type matrix_wrapper

  type :: TT_struct
     integer  ::  LL      = 0
     integer  ::  parent  = 0
     integer  ::  kids(2) = 0
  end type TT_struct

  type I1D_struct
     integer,allocatable              ::  ary(:)
  end type I1D_struct

  type I2D_struct
     integer,allocatable              ::  ary(:,:)
  end type I2D_struct

  type :: C1D_struct
     complex(MYKIND),allocatable  ::  ary(:)
  end type C1D_struct

  type :: C2D_struct
     complex(MYKIND),allocatable  ::  ary(:,:)
  end type C2D_struct

  !! HSS structure
  type :: HSSRS_struct
     type(C2D_struct),allocatable    :: D(:),B(:,:),E(:,:),UL(:),LL(:),UR(:)
     type(I1D_struct),allocatable    :: p(:,:),ipiv(:)
     integer,allocatable             :: rk(:,:),sz(:,:)

     integer,allocatable             :: crk(:,:,:) !!children rank
  end type HSSRS_struct

contains

  subroutine toep_hssrs(cv,rv,n,B,XX,nrhs,ncol,tol,maxit,tolit)
    integer,        intent(in)  :: n, ncol, nrhs
    complex(MYKIND),intent(in)  :: cv(n), rv(n), B(n,nrhs)
    real(kind=4),   intent(in)  :: tol
    integer,      optional, intent(in) :: maxit
    real(kind=4), optional, intent(in) :: tolit

    complex(MYKIND),intent(out) :: XX(n,nrhs)

    integer :: iparm(4)
    complex(MYKIND) :: BB(n,nrhs), XX2(n,nrhs), XX3(n,nrhs), c, omega, ER(n,nrhs)

    !! HSS tree and partition
    integer                      :: Lvl, lenTT
    type(TT_struct),allocatable  :: TT(:)
    integer,allocatable          :: rg(:,:)

    !! matrix A, HSS H
    type(matrix_wrapper):: A
    type(HSSRS_struct)  :: H

    !! sample data   [2,Lpar]
    type(C2D_struct) :: S(2),X(2)
    type(I1D_struct) :: indx(2) ! no use
    type(C2D_struct),allocatable :: Y(:)

    !! tmp
    integer             :: i,l,ierr, ii
    integer,allocatable :: myseed(:)
    real,allocatable    :: rbuf(:)
    real :: relres
    
    A%n=n
    allocate(A%cv(A%n),A%rv(A%n))  
    A%cv=cv
    A%rv=rv
    A%rv(1)=A%cv(1)
    call toep2cauchy(A)


    !! level of HSS tree
    Lvl = floor( log( real(n)/real(ncol*2) ) / log( 2.0 ) )

    if (Lvl<=1) then
       Lvl=2
    end if

    lenTT = 2**Lvl-1
    !write(*,*) 'Lvl',Lvl
    call TT_generate(lenTT,TT)
    call range_generate(lenTT,TT,1,n,Lvl,rg)

    
!!!!!!!!!!!!!!!!!! random seed !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !call random_seed(size=l)
    !allocate(myseed(l))
    !open(newunit=i,file="/dev/urandom",access="stream",form="unformatted",action="read",status="old")
    !read(i)myseed
    !write(*,*)"rand_seed",myseed
    !close(i)
    !call random_seed(put=myseed)
    !deallocate(myseed)

    allocate(S(1)%ary(n,ncol),S(2)%ary(n,ncol),X(1)%ary(n,ncol),X(2)%ary(n,ncol),rbuf(n*2))
    do i = 1, ncol
       call random_number(rbuf)
       X(1)%ary(:,i) = MYCMPLX(rbuf(1:n),rbuf(n+1:n*2))
    end do
    X(2)%ary = X(1)%ary

    
    !write(*,*) 'generating sampling data: S1=A X1'
    call fmatmul_hssrs('N',A,X(1)%ary,S(1)%ary,ncol)
    !write(*,*) 'generating sampling data: S2=A''X2'
    call fmatmul_hssrs('C',A,X(2)%ary,S(2)%ary,ncol)

    !! compression
    !write(*,*) 'randomized HSS compression'
    write(*,*) 'Randomized HSS construction stage'
    iparm = (/1,n,1,lenTT/)
    call HSSRS_alloc(lenTT,H)
    call loc_HSSRS_compr(iparm,A,S,X,ncol,tol,lenTT,TT,rg,indx,H%D,H%B,H%E,H%p,H%rk,H%sz)
    write(*,'(a19,i10)') "Computed HSS rank=",maxval(H%rk)

    !! HSS LU factorization
    write(*,*) '------------------------------------'
    write(*,*) 'HSS ULV factorization stage'
    call loc_HSSRS_fact(iparm,lenTT,TT,H%D,H%B,H%E,H%p,H%rk,H%sz,H%ipiv,H%UL,H%LL,H%UR)

    allocate(Y(lenTT))
    call loc_HSSRS_rhsalloc(Y,nrhs,H%sz,lenTT)

    !! solution
    write(*,*) '------------------------------------'
    write(*,*) 'HSS solution stage'

    call ifft(B,BB,n,nrhs)

    call loc_HSSRS_solL(iparm,Y,BB,nrhs,lenTT,TT,rg,H%E,H%p,H%rk,H%sz,H%ipiv,H%UL,H%LL)
    call loc_HSSRS_solU(iparm,XX,Y,nrhs,lenTT,TT,rg,H%E,H%p,H%rk,H%sz,H%UL,H%UR)

    call fft(XX,XX2,n,nrhs)

    XX=XX2/MYREAL(n)

    write(*,*) '------------------------------------'
    
    if (present(maxit) .AND. maxit>0) then
       write(*,*) 'Iterative refinement stage'
       do i=1,maxit+1
          call toepmv(BB,XX,A%cv,A%rv,n,nrhs)
          ER=B-BB

          relres=maxval(abs(ER))/maxval(abs(B))
          
          if (relres<tolit) then
             !write(*,'(a24,i2,a13)') 'Tolerance is reached at',i-1,'-th iteration.'
             exit
          end if

          if (i==maxit+1) then
             write(*,*) "Maximum number of iteration reached."
             exit
          end if

          call ifft(ER,BB,n,nrhs)

          call loc_HSSRS_solL(iparm,Y,BB,nrhs,lenTT,TT,rg,H%E,H%p,H%rk,H%sz,H%ipiv,H%UL,H%LL)
          call loc_HSSRS_solU(iparm,XX2,Y,nrhs,lenTT,TT,rg,H%E,H%p,H%rk,H%sz,H%UL,H%UR)

          call fft(XX2,XX3,n,nrhs)

          XX3=XX3/MYREAL(n)
          XX=XX+XX3
       end do
       write(*,*) 'Relative residual=',relres
             write(*,*) '------------------------------------'
    end if



    deallocate(Y)

  end subroutine toep_hssrs

  !! generate range by a top down bisection
  subroutine range_generate(lenTT,TT,head,tail,Lvl,rg)
    integer   :: lenTT,head,tail,Lvl,i,p
    type(TT_struct)      :: TT(lenTT)
    integer,allocatable :: rg(:,:)
    allocate(rg(2,lenTT))
    rg(:,lenTT)=(/head,tail/)
    do i=lenTT-1,1,-1
       if(TT(i)%LL > Lvl) cycle
       p = TT(i)%parent
       if(i==TT(p)%kids(1))then
          rg(:,i)=(/rg(1,p), (rg(1,p)+rg(2,p))/2  /)
       else
          rg(:,i)=(/(rg(1,p)+rg(2,p))/2+1, rg(2,p)/)
       end if
    end do
  end subroutine range_generate


!!!!!!!!!!!!!!!!!!!!!!  1. elimination tree generation  !!!!!!!!!!!!!!!!!!!
  subroutine TT_generate( lenTT, TT )

    !! This function is only valid for lenTT == 2**k-1 or lenTT == 2**k+1
    integer                       ::  lenTT
    type(TT_struct),allocatable   ::  TT(:)
    integer                       ::  L1, L2, level, ind, Ntmp
    integer,allocatable           ::  nop(:,:)

    ! level bounded by L1 and L2: DO NOT use floor and ceiling, not numerically stable
    L1 = int(log(real(lenTT+1))/log(2.0))
    L2 = L1+1
    if (lenTT+1 == 2**L1) then
       L2 = L1
    elseif (lenTT+1 == 2**L2) then
       L1 = L2
    end if

    allocate(TT(lenTT))

    ! building up TT
    allocate(nop(L2,3))
    nop = 0
    do Ntmp = 1,lenTT

       !! with Schur complement as additional block
       if (Ntmp == lenTT-1 .and. L1 /= L2) then    
          TT(lenTT-2)%parent = lenTT
          TT(lenTT-1)%kids   = 0
          TT(lenTT-1)%parent = lenTT
          TT(lenTT)%kids(1)  = lenTT-2
          TT(lenTT)%kids(2)  = lenTT-1
          exit    ! IMPORTANT
       end if

       ind = 0
       do level = 1,L2
          if (nop(level,1) == 2) then
             ind = level    ! search for the level which has two no-parent nodes
             exit
          end if
       end do

       if (ind == 0) then    ! starting from the bottom level
          level = L2
          nop(level,1) = nop(level,1) + 1
          if ( nop(level,2)==0 ) then
             nop(level,2) = Ntmp
          else
             nop(level,3) = Ntmp
          end if
       else
          level = ind - 1    ! go to the upper level to form the tree
          nop(level,1) = nop(level,1) + 1
          if ( nop(level,2) == 0 ) then
             nop(level,2) = Ntmp
          else
             nop(level,3) = Ntmp
          end if
          TT(Ntmp)%kids = nop(ind,2:3)
          TT( nop(ind,2) )%parent = Ntmp
          TT( nop(ind,3) )%parent = Ntmp
          nop(ind,:) = 0    ! clear step: start over
       end if

    end do
    deallocate(nop)

    ! set up level information
    TT(lenTT)%LL = 1
    do Ntmp = lenTT-1,1,-1
       TT(Ntmp)%LL = TT( TT(Ntmp)%parent )%LL + 1
    end do

  end subroutine TT_generate


  subroutine HSSRS_alloc(length,A)
    integer             :: length
    type(HSSRS_struct)    :: A
    allocate(A%p(2,length),A%rk(2,length),A%sz(2,length),A%E(2,length),A%B(2,length))
    allocate(A%UL(length),A%LL(length),A%UR(length),A%ipiv(length),A%D(length))
  end subroutine HSSRS_alloc


  subroutine loc_HSSRS_compr(locparm,A,S,X,ncol,tol,lenTT,TT,rg,indx,D,B,E,p,rk,sz)
    integer :: locparm(4)
    integer              :: ncol,lenTT
    type(matrix_wrapper) :: A
    real(4)              :: tol
    type(TT_struct)      :: TT(lenTT)
    integer              :: rg(2,lenTT)
    type(C2D_struct)     :: S(2),X(2) !sample data

    !!----------------output------------------------
    type(I1D_struct)     :: indx(2)   !index of B_rt
    !! HSS
    type(C2D_struct) :: D(lenTT),B(2,lenTT),E(2,lenTT)
    type(I1D_struct) :: p(2,lenTT)
    integer          :: rk(2,lenTT),sz(2,lenTT)

    !!-----------------temp-------------------------
    integer              :: i,c1,c2,sz1,sz2,k,head,mid,lwork,info,n
    integer              :: order(locparm(2),2)
    complex(MYKIND)       :: M(ncol,locparm(2),2),tau(ncol)
    complex(MYKIND)       :: work(locparm(2)+1)
    real(MYKIND)          ::rwork(locparm(2)*2),R11

    !! diagonal blocks
    call fgetD(A,D,lenTT,TT,rg)

    !! no off-diag return
    if(lenTT == 1.and.locparm(3)==1) then
       !write(*,*) 'no local HSS'
       sz(:,1)=locparm(2)
       rk(:,1)=0
       return
    end if

    !! size,minimum rank 1,initial order
    n = locparm(2); lwork = n + 1
    rk = 1
    do i = 1,n
       order(i,:)=rg(1,1)+i-1
    end do

    do i = 1,lenTT
       !! 1st row of the current block
       head = rg(1,i)-rg(1,1)+1

       if(TT(i)%kids(1)==0) then
          sz1 = rg(2,i)-rg(1,i)+1; sz2 = sz1

          !! S-DX
          call GEMM('N','N',sz1,ncol,sz1,-ONE,D(i)%ary,sz1,X(1)%ary(head,1),n,ONE,S(1)%ary(head,1),n)
          call GEMM('C','N',sz1,ncol,sz1,-ONE,D(i)%ary,sz1,X(2)%ary(head,1),n,ONE,S(2)%ary(head,1),n)
       else
          c1 = TT(i)%kids(1); c2 = TT(i)%kids(2)
          sz1 = rk(1,c1)+rk(1,c2); sz2 = rk(2,c1)+rk(2,c2)

          !! 1st row of c2
          mid = rg(1,c2)-rg(1,1)+1
          allocate(B(1,i)%ary(rk(1,c1),rk(2,c2)))
          allocate(B(2,i)%ary(rk(1,c2),rk(2,c1)))

          !! get B
          call fgetB(A,B(1,i)%ary,order(head,1),rk(1,c1),order(mid ,2),rk(2,c2))
          call fgetB(A,B(2,i)%ary,order(mid ,1),rk(1,c2),order(head,2),rk(2,c1))

          !! update sample data
          ! S1_c1 - B_c1*VX1_c2; S1_c2 - B_c2*VX1_c1; S2_c1 - B_c2'*VX2_c2; S2_c2 - B_c1'*VX2_c1
          call GEMM('N','N',rk(1,c1), ncol, rk(2,c2), -ONE, B(1,i)%ary, rk(1,c1),&
               X(1)%ary(mid ,1), n, ONE, S(1)%ary(head,1), n)
          call GEMM('N','N',rk(1,c2), ncol, rk(2,c1), -ONE, B(2,i)%ary, rk(1,c2),&
               X(1)%ary(head,1), n, ONE, S(1)%ary(mid ,1), n)
          call GEMM('C','N',rk(2,c1), ncol, rk(1,c2), -ONE, B(2,i)%ary, rk(1,c2),&
               X(2)%ary(mid ,1), n, ONE, S(2)%ary(head,1), n)
          call GEMM('C','N',rk(2,c2), ncol, rk(1,c1), -ONE, B(1,i)%ary, rk(1,c1),&
               X(2)%ary(head,1), n, ONE, S(2)%ary(mid ,1), n)

          !! uplift S,X,order from c2
          S(1)%ary(head+rk(1,c1):head+sz1-1,:)=S(1)%ary(mid:mid+rk(1,c2)-1,:)
          X(2)%ary(head+rk(1,c1):head+sz1-1,:)=X(2)%ary(mid:mid+rk(1,c2)-1,:)
          order(head+rk(1,c1):head+sz1-1,1)=   order(mid:mid+rk(1,c2)-1,1)

          S(2)%ary(head+rk(2,c1):head+sz2-1,:)=S(2)%ary(mid:mid+rk(2,c2)-1,:)
          X(1)%ary(head+rk(2,c1):head+sz2-1,:)=X(1)%ary(mid:mid+rk(2,c2)-1,:)
          order(head+rk(2,c1):head+sz2-1,2)=   order(mid:mid+rk(2,c2)-1,2)
       end if

       sz(:,i) = (/sz1,sz2/)
       !! no off-diag left return
       if(i == lenTT.and.locparm(3)==1) then
          !write(*,*) 'no indx, local root PIE'
          rk(:,i) = 0
          return
       end if

       allocate(p(1,i)%ary(sz1),p(2,i)%ary(sz2))
       p(1,i)%ary = 0; p(2,i)%ary = 0

       !! transpose sampling data: AX
       M(:,1:sz1,1)=transpose(S(1)%ary(head:sz1+head-1,:))
       M(:,1:sz2,2)=transpose(S(2)%ary(head:sz2+head-1,:))

       !! RRQR cGEQP3(m,n,a,lda,jpvt,tau,work,lwork,rwork,info)
       call GEQP3(ncol,sz1,M(1,1,1),ncol,p(1,i)%ary,tau,work,lwork,rwork,info)
       call GEQP3(ncol,sz2,M(1,1,2),ncol,p(2,i)%ary,tau,work,lwork,rwork,info)

       !! truncate rank
       R11=abs(M(1,1,1))
       if(R11 > 0) then
          do k = 2,min(sz1,ncol)
             if(abs(M(k,k,1))<tol*R11) exit
             rk(1,i)=k
          end do
       end if
       R11=abs(M(1,1,2))
       if(R11 > 0) then
          do k = 2,min(sz2,ncol)
             if(abs(M(k,k,2))<tol*R11) exit
             rk(2,i)=k
          end do
       end if

       !! E
       if(locparm(1) == 0) then
          k = rk(1,i)
       else
          k = max(rk(1,i),rk(2,i))
       end if
       if (k<sz1) then
          allocate(E(1,i)%ary(sz1-k,k)); E(1,i)%ary = ZERO
          !!R1\R2
          call TRTRS('U','N','N',rk(1,i),sz1-k,M(1,1,1),ncol,M(1,k+1,1),ncol,info)
          E(1,i)%ary(:,1:rk(1,i))=transpose(M(1:rk(1,i),k+1:sz1,1))
       end if
       rk(1,i) = k
       if(locparm(1) == 0) then
          k = rk(2,i)
       else
          k = max(rk(1,i),rk(2,i))
       end if
       if (k<sz2) then
          allocate(E(2,i)%ary(sz2-k,k)); E(2,i)%ary = ZERO
          call TRTRS('U','N','N',rk(2,i),sz2-k,M(1,1,2),ncol,M(1,k+1,2),ncol,info)
          E(2,i)%ary(:,1:rk(2,i))=transpose(M(1:rk(2,i),k+1:sz2,2))
       end if
       rk(2,i) = k

       !![I,Ei']Pi'Xj
       k = rk(2,i)
       if (k<sz2) then
          X(1)%ary(head:head+sz2-1,:) = X(1)%ary(p(2,i)%ary+(head-1),:)
          call GEMM('C','N',k,ncol,sz2-k,ONE,E(2,i)%ary,sz2-k,X(1)%ary(head+k,1),n,ONE,X(1)%ary(head,1),n)
          S(2)%ary(head:head+k-1,:)   = S(2)%ary(p(2,i)%ary(1:k)+(head-1),:)
          order(head:head+k-1,2)   =    order(p(2,i)%ary(1:k)+(Head-1),2)
       else
          deallocate(p(2,i)%ary)
       end if
       k = rk(1,i)
       if (k<sz1) then 
          X(2)%ary(head:head+sz1-1,:)=X(2)%ary(p(1,i)%ary+(head-1),:)
          call GEMM('C','N',k,ncol,sz1-k,ONE,E(1,i)%ary,sz1-k,X(2)%ary(head+k,1),n,ONE,X(2)%ary(head,1),n)
          S(1)%ary(head:head+k-1,:)  =S(1)%ary(p(1,i)%ary(1:k)+(head-1),:)
          order(head:head+k-1,1)   =  order(p(1,i)%ary(1:k)+(Head-1),1)
       else
          deallocate(p(1,i)%ary)
       end if
       !write(*,*)'i',i,'rk',rk(:,i),'sz',sz(:,i)
    end do
    allocate(indx(1)%ary(rk(1,lenTT)),indx(2)%ary(rk(2,lenTT)))
    indx(1)%ary = order(1:rk(1,lenTT),1)
    indx(2)%ary = order(1:rk(2,lenTT),2)
  end subroutine loc_HSSRS_compr


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  !!                local HSS randomized LU factorization
  !! locparm(1): 0 unsymmetric pattern, 1 symmetric pattern
  !! locparm(2): nrow (not used)
  !! locparm(3)
  !! 0 nonstop: return indx of B, S,X holds reduced sample data
  !! 1    stop: no indx
  !! locparm(4)
  !! 0 incomplete: no full LU
  !! npv         : full LU at npv, stores at root
  !! lenTT       : full LU at root
  subroutine loc_HSSRS_fact(locparm,lenTT,TT,D,B,E,p,rk,sz,ipiv,UL,LL,UR)
    !!-----------------input------------------------
    integer          :: locparm(4),lenTT
    type(TT_struct)  :: TT(lenTT)
    type(C2D_struct) :: D(lenTT),B(2,lenTT),E(2,lenTT)
    type(I1D_struct) :: p(2,lenTT)
    integer          :: rk(2,lenTT),sz(2,lenTT)
    !! factor
    type(C2D_struct) :: UL(lenTT),LL(lenTT),UR(lenTT)
    type(I1D_struct) :: ipiv(lenTT)
    integer :: i,itail,c1,c2,l1,l2,ld,info

    if(locparm(1)==0) then
       write(*,*)"rank pattern unsymmetric!"
       return
    end if

    if(locparm(4)==0) then
       itail = lenTT
    else
       itail = locparm(4)
    end if

    do i=1,itail
       ld = sz(1,i)
       if(TT(i)%kids(1)/=0) then
          c1 = TT(i)%kids(1); c2 = TT(i)%kids(2)
          l1 = rk(1,c1); l2 = rk(1,c2)
          allocate(D(i)%ary(ld,ld))

          !! merge D,B
          D(i)%ary(1:l1,1:l1) = D(c1)%ary(1:l1,1:l1)
          D(i)%ary(l1+1:ld,l1+1:ld) = D(c2)%ary(1:l2,1:l2)
          D(i)%ary(1:l1,l1+1:ld) = B(1,i)%ary
          D(i)%ary(l1+1:ld,1:l1) = B(2,i)%ary
          deallocate(D(c1)%ary,D(c2)%ary,B(1,i)%ary,B(2,i)%ary)
       end if

       l1 = rk(1,i); l2 = sz(1,i)-rk(1,i)
       if(min(l1,l2) > 0) then
          !! P1'DP2
          D(i)%ary=D(i)%ary(p(1,i)%ary,p(2,i)%ary)
          !![I,O;-E1;I]D[I,-E2';O,I]
          call GEMM('N','N',l2,ld,l1,-ONE,E(1,i)%ary,l2,D(i)%ary,ld,ONE,D(i)%ary(l1+1,1),ld)
          call GEMM('N','C',ld,l2,l1,-ONE,D(i)%ary,ld,E(2,i)%ary,l2,ONE,D(i)%ary(1,l1+1),ld)

          allocate(ipiv(i)%ary(l2))
          allocate(LL(i)%ary(l2,l1),UL(i)%ary(l2,l2),UR(i)%ary(l2,l1))

          !! D -> UL,UR,LL
          UL(i)%ary = D(i)%ary(l1+1:ld,l1+1:ld)
          UR(i)%ary = D(i)%ary(l1+1:ld,1:l1)
          LL(i)%ary = transpose(D(i)%ary(1:l1,l1+1:ld))

          !![P'L 0; LL^T I][U UR; 0 S]
          call GETRF(l2,l2,UL(i)%ary,l2,ipiv(i)%ary,info)
          call LASWP(l1,UR(i)%ary,l2,1,l2,ipiv(i)%ary,1)
          call TRTRS('L','N','U',l2,l1,UL(i)%ary,l2,UR(i)%ary,l2,info)
          call TRTRS('U','T','N',l2,l1,UL(i)%ary,l2,LL(i)%ary,l2,info)
          call GEMM('T','N',l1,l1,l2,-ONE,LL(i)%ary,l2,UR(i)%ary,l2,ONE,D(i)%ary,ld)
       end if
    end do
    if(locparm(4)==0) return !return to parallel subroutine

    !! only used in sequential factorization
    if(locparm(4)==lenTT) then
       i = lenTT; ld = sz(1,i)
       allocate(UL(i)%ary(ld,ld),ipiv(i)%ary(ld))
       UL(i)%ary=D(i)%ary
       call GETRF(ld,ld,UL(i)%ary,ld,ipiv(i)%ary,info)
       deallocate(D(i)%ary)
    else
       i = lenTT
       c1 = TT(i)%kids(1); c2  = TT(i)%kids(2)
       l1 = rk(1,c1); l2 = rk(1,c2)

       !! factorize reduced D, update Schur omplement
       allocate(UL(i)%ary(l1,l1),ipiv(i)%ary(l1))
       UL(i)%ary=D(c1)%ary(1:l1,1:l1)
       deallocate(D(c1)%ary)
       call GETRF(l1,l1,UL(i)%ary,l1,ipiv(i)%ary,info)

       !! UR: L\B1[I;E2']P2'; LL: U\B2^T[I;E1^T]P1^T

       ld = sz(1,c2)
       allocate(UR(i)%ary(l1,ld),LL(i)%ary(l1,ld))
       UR(i)%ary(:,1:l2)=B(1,i)%ary
       LL(i)%ary(:,1:l2)=transpose(B(2,i)%ary)
       call LASWP(l2,UR(i)%ary,l1,1,l1,ipiv(i)%ary,1)
       call TRTRS('L','N','U',l1,l2,UL(i)%ary,l1,UR(i)%ary,l1,info)
       call TRTRS('U','T','N',l1,l2,UL(i)%ary,l1,LL(i)%ary,l1,info)

       call GEMM('N','C',l1,ld-l2,l2,ONE,UR(i)%ary,l1,E(2,c2)%ary,ld-l2,ZERO,UR(i)%ary(1,l2+1),l1)
       call GEMM('N','T',l1,ld-l2,l2,ONE,LL(i)%ary,l1,E(1,c2)%ary,ld-l2,ZERO,LL(i)%ary(1,l2+1),l1)
       UR(i)%ary(:,p(2,c2)%ary)=UR(i)%ary
       LL(i)%ary(:,p(1,c2)%ary)=LL(i)%ary

       call GEMM('T','N',ld,ld,l1,-ONE,LL(i)%ary,l1,UR(i)%ary,l1,ONE,D(c2)%ary,ld)
    end if
  end subroutine loc_HSSRS_fact

  subroutine loc_HSSRS_rhsalloc(Y,nrhs,sz,lenTT)
    integer          :: lenTT,nrhs,i
    type(C2D_struct) :: Y(lenTT)
    integer          :: sz(2,lenTT)

    do i = 1,lenTT
       allocate(Y(i)%ary(sz(1,i),nrhs))
    end do
  end subroutine loc_HSSRS_rhsalloc

  subroutine loc_HSSRS_solL(locparm,Y,B,nrhs,lenTT,TT,rg,E,p,rk,sz,ipiv,UL,LL)
    integer          :: locparm(4),lenTT,nrhs
    type(TT_struct)  :: TT(lenTT)
    integer          :: rg(2,lenTT)

    type(C2D_struct) :: E(2,lenTT),UL(lenTT),LL(lenTT)
    type(I1D_struct) :: p(2,lenTT)
    integer          :: rk(2,lenTT),sz(2,lenTT)
    type(I1D_struct) :: ipiv(lenTT)

    complex(MYKIND)   :: B(locparm(2),nrhs)
    type(C2D_struct) :: Y(lenTT)

    integer :: par,i,l1,l2,ld,info,itail,tail,c1,c2

    if(locparm(4)==0) then
       itail = lenTT
    else
       itail = locparm(4)
    end if

    do i = 1,itail
       ld = sz(1,i)
       if(TT(i)%kids(1)==0) then
          Y(i)%ary=B(rg(1,i)-rg(1,1)+1:rg(2,i)-rg(1,1)+1,:)
       else
          c1 = TT(i)%kids(1); c2 = TT(i)%kids(2)
          l1 = rk(1,c1)
          Y(i)%ary(1:l1,:)    = Y(c1)%ary(1:l1,:)
          Y(i)%ary(l1+1:ld,:) = Y(c2)%ary(1:ld-l1,:)
       end if


       l1 = rk(1,i); l2 = sz(1,i)-rk(1,i)
       if(min(l1,l2) > 0) then
          Y(i)%ary=Y(i)%ary(p(1,i)%ary,:)
          call GEMM('N','N',l2,nrhs,l1,-ONE,E(1,i)%ary,l2,Y(i)%ary,ld,ONE,Y(i)%ary(l1+1,1),ld)
          call LASWP(nrhs,Y(i)%ary(l1+1,1),ld,1,l2,ipiv(i)%ary,1)
          call TRTRS('L','N','U',l2,nrhs,UL(i)%ary,l2,Y(i)%ary(l1+1,1),ld,info)
          call GEMM('T','N',l1,nrhs,l2,-ONE,LL(i)%ary,l2,Y(i)%ary(l1+1,1),ld,ONE,Y(i)%ary,ld)               
       end if
    end do

    if(locparm(4)==0) return !return to parallel subroutine

    !! only used in sequential solution
    if(locparm(4)==lenTT) then
       i = lenTT; ld = sz(1,i)

       call LASWP(nrhs,Y(i)%ary,ld,1,ld,ipiv(i)%ary,1)
       call TRTRS('L','N','U',ld,nrhs,UL(i)%ary,ld,Y(i)%ary,ld,info)
    else
       i = lenTT
       c1 = TT(i)%kids(1); c2  = TT(i)%kids(2)
       l1 = rk(1,c1); l2 = sz(1,c2); ld = sz(1,c1)

       call LASWP(nrhs,Y(c1)%ary,ld,1,l1,ipiv(i)%ary,1)
       call TRTRS('L','N','U',l1,nrhs,UL(i)%ary,l1,Y(c1)%ary,ld,info)

       !! Yc2=Yc2-LL^TYc1
       call GEMM('T','N',l2,nrhs,l1,-ONE,LL(i)%ary,l1,Y(c1)%ary,ld,ONE,Y(c2)%ary,l2)
    end if
  end subroutine loc_HSSRS_solL

  subroutine loc_HSSRS_solU(locparm,X,Y,nrhs,lenTT,TT,rg,E,p,rk,sz,UL,UR)
    integer          :: locparm(4),lenTT,nrhs
    type(TT_struct)  :: TT(lenTT)
    integer          :: rg(2,lenTT)

    type(C2D_struct) :: E(2,lenTT),UL(lenTT),UR(lenTT)
    type(I1D_struct) :: p(2,lenTT)
    integer          :: rk(2,lenTT),sz(2,lenTT)

    complex(MYKIND)    :: X(locparm(2),nrhs)
    type(C2D_struct)  :: Y(lenTT)

    integer :: par,i,l1,l2,ld,info,tail,c1,c2,ihead

    ihead=lenTT

    !! only used in sequential solution
    if(locparm(4)/=0) then
       if(locparm(4)==lenTT) then
          i = lenTT; ld = sz(1,i)
          call TRTRS('U','N','N',ld,nrhs,UL(i)%ary,ld,Y(i)%ary,ld,info)
       else
          i = lenTT
          c1 = TT(i)%kids(1); c2 = TT(i)%kids(2)
          l1 = rk(1,c1); l2  = sz(1,c2); ld = sz(1,c1)

          call GEMM('N','N',l1,nrhs,l2,-ONE,UR(i)%ary,l1,Y(c2)%ary,l2,ONE,Y(c1)%ary,ld)

          call TRTRS('U','N','N',l1,nrhs,UL(i)%ary,l1,Y(c1)%ary,ld,info)
          ihead=c1
       end if
    end if

    do i = ihead,1,-1
       ld = sz(1,i)
       l1 = rk(1,i); l2 = sz(1,i)-rk(1,i)
       if(min(l1,l2) > 0) then
          ![I,0;UR,UL]\Y
          call GEMM('N','N',l2,nrhs,l1,-ONE,UR(i)%ary,l2,Y(i)%ary,ld,ONE,Y(i)%ary(l1+1,1),ld)
          call TRTRS('U','N','N',l2,nrhs,UL(i)%ary,l2,Y(i)%ary(l1+1,1),ld,info)

          call GEMM('C','N',l1,nrhs,l2,-ONE,E(2,i)%ary,l2,Y(i)%ary(l1+1,1),ld,ONE,Y(i)%ary,ld)
          Y(i)%ary(p(2,i)%ary,:)=Y(i)%ary
       end if
       if(TT(i)%kids(1)==0) then
          X(rg(1,i)-rg(1,1)+1:rg(2,i)-rg(1,1)+1,:)=Y(i)%ary
       else
          c1 = TT(i)%kids(1); c2 = TT(i)%kids(2)
          l1 = rk(1,c1)
          Y(c1)%ary(1:l1,:) = Y(i)%ary(1:l1,:)
          Y(c2)%ary(1:ld-l1,:) = Y(i)%ary(l1+1:ld,:)
       end if
    end do
  end subroutine loc_HSSRS_solU


  subroutine fgetD(A,D,lenTT,TT,rg)
    type(matrix_wrapper),intent(in) :: A
    integer,             intent(in) :: lenTT
    type(TT_struct),     intent(in) :: TT(lenTT)
    integer,             intent(in) :: rg(2,lenTT)

    type(C2D_struct),   intent(out) :: D(lenTT)

    integer :: ii, sz, index
    integer,allocatable :: I(:)

    do ii=1,lenTT
       if (TT(ii)%kids(1)==0) then
          sz=rg(2,ii)-rg(1,ii)+1
          allocate(D(ii)%ary(sz,sz),I(sz))
          I=(/ ( index, index=rg(1,ii),rg(2,ii)  )/)
          call cauchyij(D(ii)%ary,I,sz,I,sz,A)
          deallocate(I)
       end if
    end do

  end subroutine fgetD


  subroutine fgetB(A,B,I,k,J,l)
    type(matrix_wrapper),intent(in) :: A
    integer,             intent(in) :: k, l
    integer,             intent(in) :: I(k), J(l)

    complex(MYKIND),    intent(out) :: B(k,l)

    call cauchyij(B,I,k,J,l,A)

  end subroutine fgetB


  subroutine cauchyij(M,I,k,J,l,A)
    integer,              intent(in) :: k, l
    integer,              intent(in) :: I(k), J(l)
    type(matrix_wrapper), intent(in) :: A

    complex(MYKIND),intent(out):: M(k,l)

    integer :: ii, jj
    complex(MYKIND) ::  w

    do jj=1,l
       do ii=1,k
          if (I(ii)==J(jj)) then
             M(ii,jj)=A%x(I(ii))
          else
             M(ii,jj)=A%u(I(ii),1)*A%v(1,J(jj))+A%u(I(ii),2)*A%v(2,J(jj))
             !write(*,*) M(ii,jj)
             w=A%d(I(ii))-A%d(J(jj))
             !write(*,*) A%d(I(ii)),A%d(J(jj)),w
             M(ii,jj)=M(ii,jj)/w
          end if
          !write(*,*) I(ii),J(jj),M(ii,jj)
       end do
    end do
  end subroutine cauchyij

  
  subroutine toep2cauchy(A)
    type(matrix_wrapper) :: A

    integer         :: i
    complex(MYKIND) :: uu(A%n,2), vv(A%n,2), omega, x(A%n), d(A%n)

    allocate(A%u(A%n,2),A%v(2,A%n),A%x(A%n),A%d(A%n))

    A%u=0
    A%v=0

    A%u(1,1)=ONE
    A%v(2,A%n)=ONE

    do i=2,A%n
       A%u(i,2)=A%rv(A%n+2-i)-A%cv(i)
    end do

    do i=1,A%n-1
       A%v(1,i)=-A%u(A%n+1-i,2)
    end do

    call ifft(A%u,uu,A%n,2)
    !A%u=uu/sqrt(MYREAL(A%n))
    A%u=uu/MYREAL(A%n)

    call fft(transpose(A%v),vv,A%n,2)

    A%v=transpose(vv)

    omega=MYCMPLX(cos(2*PI/MYREAL(A%n)),sin(2*PI/MYREAL(A%n)))
    !write(*,*) '---theta---',PI,2/MYREAL(A%n)*PI
    !write(*,*) '---omega---',omega
    !write(*,*) 'cos(pi)',cos(PI),'cos(2pi)',cos(2*PI)
    d=ONE
    do i=2,A%n
       d(i)=omega*d(i-1)
    end do
    A%d=d
    
    x=A%cv(1)
    do i=2,A%n
       x(i)=(i-1)*A%rv(A%n+2-i)+(A%n+1-i)*A%cv(i)
       x(i)=x(i)/MYREAL(A%n)
    end do

    call fft(x,A%x,A%n,1)
    x=A%x
    A%x(1)=x(1)
    do i=2,A%n
       A%x(i)=x(A%n+2-i)
    end do

    !write(*,*) '---u---',A%u
    !write(*,*) '---v---',A%v
    !write(*,*) '---d---',A%d
    !write(*,*) '---x---',A%x
  end subroutine toep2cauchy


  subroutine toepmv(y,x,cv,rv,n,m)
    ! toeplitz mat-vec multiplication y=T*x

    integer, intent(in) :: n,m
    complex(MYKIND),intent(in) :: cv(n), rv(n), x(n,m)
    complex(MYKIND),intent(out) :: y(n,m)

    complex(MYKIND) :: c(2*n),c1(2*n)
    complex(MYKIND) :: xt(2*n,m),xtt(2*n,m)
    integer :: i
    !real(MYKIND) :: l

    ! column wins on cv(1)
    c(1:n)=cv
    c(n+1)=0
    c((n+2):(2*n))=rv(n:2:-1)

    xt(1:n,:)=x
    xt((n+1):(2*n),:)=0

    call ifft(xt,xtt,2*n,m)
    call ifft(c,c1,2*n,1)

    do i=1,2*n
       xtt(i,:)=xtt(i,:)*c1(i)
    end do

    xt=xtt

    call fft(xt,xtt,2*n,m)

    y=xtt(1:n,:)

    y=y/MYREAL(2*n)

  end subroutine toepmv


  subroutine fft(x,y,a,b)
    ! y=fft(x): non-normalized fft on each column of x

    integer,intent(in) :: a,b
    complex(MYKIND),intent(in)  :: x(a,b)
    complex(MYKIND),intent(out) :: y(a,b)

    integer*8 :: plan
    integer :: rank = 1
    integer :: n, howmany, idist, odist, inembed, onembed
    integer :: istride = 1, ostride = 1 ! distance between two elements in the same column

    n=a
    howmany=b

    idist=n
    odist=n ! the distance in memory between the first element of the first array and the first element of the second array

    inembed=n
    onembed=n

    call PLAN_MANY(plan,rank,n,howmany,x,inembed,istride,idist,y,onembed,ostride,odist,-1,FFTW_ESTIMATE)
    call EXECUTE(plan)
    call DESTROY(plan)

  end subroutine fft


  subroutine ifft(x,y,a,b)
    ! y=ifft(x): non-normalized ifft on each column of x

    integer,        intent(in)  :: a,b
    complex(MYKIND),intent(in)  :: x(a,b)
    complex(MYKIND),intent(out) :: y(a,b)

    integer*8 :: plan
    integer   :: rank = 1
    integer   :: n, howmany, idist, odist, inembed, onembed
    integer   :: istride = 1, ostride = 1

    n=a
    howmany=b

    idist=n
    odist=n
    inembed=n
    onembed=n

    call PLAN_MANY(plan,rank,n,howmany,x,inembed,istride,idist,y,onembed,ostride,odist,1,FFTW_ESTIMATE)
    call EXECUTE(plan)
    call DESTROY(plan)

  end subroutine ifft


  subroutine fmatmul_hssrs(trans,A,x,y,nrhs)
    ! compute y=C*x
    ! C : corresponding Cauchy-like matrix w.r.t. T=Toeplitz(rv,cv)

    character,intent(in) :: trans
    type(matrix_wrapper) :: A
    integer,intent(in)   :: nrhs
    complex(MYKIND),intent(in) :: x(A%n,nrhs)
    complex(MYKIND),intent(out):: y(A%n,nrhs)

    complex(MYKIND) :: s, omega, yt(A%n,nrhs)
    integer :: j

    call fft(x,y,A%n,nrhs)
    !y=y/sqrt(MYREAL(A%n))

    if (trans=='N') then
       call toepmv(yt,y,A%cv,A%rv,A%n,nrhs)
    else
       call toepmv(yt,y,conjg(A%rv),conjg(A%cv),A%n,nrhs)
    end if

    call ifft(yt,y,A%n,nrhs)
    !y=y/sqrt(MYREAL(A%n))

    y=y/MYREAL(A%n)

  end subroutine fmatmul_hssrs

end module toephssrs
